import json
import os.path
from pathlib import Path
import sys
import time
import shutil
from contextlib import contextmanager

import click
from flask import abort, Flask, jsonify, request
from flask_cors import cross_origin
from markupsafe import escape

from . import default_settings, initdb, scorer
from .db import get_db, getCSVCols
from .extend import processDataExtensionBatch
from .preview import getEntity
from .score import processQueryBatch

try:
    import importlib_metadata as metadata
except:
    from importlib import metadata

__version__ = '0.3.2'
#------------------------------------------------------------------
# Implement reconciliation API
# [[https://reconciliation-api.github.io/specs/latest/]]
#------------------------------------------------------------------


@contextmanager
def Timer():
    t = time.perf_counter()
    print("start timer", flush=True)
    yield
    elapsed = time.perf_counter() - t
    print("Elapsed: %s" % (elapsed,))


# Default manifest.  Can be overriden/updated in configuration
MANIFEST = {
    "versions": ["0.1"],
    "name": "CSV Reconcile",
    "identifierSpace": "http://localhost/csv_reconcile/ids",
    "schemaSpace": "http://localhost/csv_reconcile/schema",
    "extend": {
        "propose_properties": {
            "service_url": "http://localhost:5000",
            "service_path": "/properties"
        }
    }
}


def create_app(config=None, instance_path=None, scorerOption=None):
    app = Flask("csv-reconcile", instance_path=instance_path)

    instance_path = Path(app.instance_path)

    try:
        os.makedirs(instance_path)
    except OSError:
        pass

    scorerfile = instance_path / 'scorer.txt'

    # clean up old files if they exist
    # "" indicates called from doinit()
    if scorerOption == "" and scorerfile.is_file():
        scorerfile.unlink()
    elif scorerOption:
        with open(scorerfile, 'w') as f:
            f.write(scorerOption)

    scorerOption = None
    if scorerfile.is_file():
        with open(scorerfile) as f:
            scorerOption = f.read()

    if pickScorer(scorerOption) is None:
        return None

    # possibly better to roll THRESHOLD and LIMIT into one config called LIMITS
    app.config.from_object(default_settings)

    cfgfile = instance_path / "reconcile.config"

    # clean up old configs if they exist
    # "" indicates called from doinit()
    if config == "" and cfgfile.is_file():
        cfgfile.unlink()
    elif config:
        shutil.copyfile( config, cfgfile )

    if cfgfile.is_file():
        app.config.from_pyfile(cfgfile)

    scoreOptions = app.config['SCOREOPTIONS']
    scorer.processScoreOptions(scoreOptions)

    if 'MANIFEST' in app.config:
        MANIFEST.update(app.config['MANIFEST'])

    loglevel = app.config['LOGLEVEL']
    if loglevel:
        app.logger.setLevel(loglevel)

    @app.before_request
    def before():
        app.logger.debug(request.method)
        app.logger.debug(request.headers)

    @app.after_request
    def after(response):
        app.logger.debug(response.headers)
        return response

    @app.route('/reconcile', methods=['POST', 'GET'])
    @cross_origin()
    def acceptQuery():
        threshold = app.config.get('THRESHOLD', None)
        limit = app.config.get('LIMIT', None)
        scoreOptions = app.config['SCOREOPTIONS']
        queries = request.form.get('queries')
        extend = request.form.get('extend')
        if queries:
            db = get_db()

            queryBatch = json.loads(queries)

            app.logger.info(queryBatch)
            with Timer():
                ret = processQueryBatch(db,
                                        queryBatch,
                                        limit=limit,
                                        threshold=threshold,
                                        **scoreOptions)
            app.logger.info(ret)
            return ret
        elif extend:
            extendBatch = json.loads(extend)

            app.logger.info(extendBatch)
            with Timer():
                ret = processDataExtensionBatch(extendBatch)
            app.logger.info(ret)
            return ret
        else:
            return MANIFEST

    # FIX FIX FIX...  Not needed in OpenRefine 3.5
    # [[https://github.com/OpenRefine/OpenRefine/issues/3672]]
    def jsonpify(obj):
        """
        Like jsonify but wraps result in a JSONP callback if a 'callback'
        query param is supplied.
        """
        try:
            callback = request.args['callback']
            response = app.make_response("%s(%s)" % (callback, json.dumps(obj)))
            response.mimetype = "text/javascript"
            return response
        except KeyError:
            return jsonify(obj)

    @app.route('/properties', methods=['POST', 'GET'])
    @cross_origin()
    def acceptPropertyRequest():
        # query string arg
        propType = request.args.get('type')

        # Type irrelevant, return all columns
        if propType != None:
            cols = getCSVCols()
            ret = dict(properties=[{
                'id': colname,
                'name': name
            } for colname, name in cols])
            return jsonpify(ret)

        # unprocessible request

    @app.route('/preview/<entity_id>')
    @cross_origin()
    def preview_service(entity_id=None):
        if not entity_id:
            abort(404)
        entity = getEntity(entity_id)
        if not entity:
            abort(404)
        entity_html = "".join([f"<dt>{escape(key)}</dt><dd>{escape(val)}</dd>"
                               for key, val in entity.items()])
        return f"""<!DOCTYPE html>
<html>
    <head>
        <meta charset='utf-8'>
        <title>Preview for {escape(entity_id)}</title>
        <style type='text/css'>
          h1 {{font-size: 115%; }}
          dl {{display: flex; flex-flow: row wrap;}}
          dt {{flex-basis: 20%; padding: 2px 4px;
               text-align: right; font-weight: bold;}}
          dt::after {{content: ':';}}
          dd {{flex-basis: 70%; flex-grow: 1;
               margin: 0; padding: 2px 4px;}}
        </style>
    </head>
    <body>
        <dl>{entity_html}</dl>
    </body>
</html>"""

    return app


def pickScorer(plugin):
    eps = metadata.entry_points().select(group='csv_reconcile.scorers')
    entrypoint = None
    if len(eps) == 0:
        raise RuntimeError("Please install a \"csv_reconcile.scorers\" plugin")
    elif plugin:
        for ep in eps:
            if ep.name == plugin:
                entrypoint = ep
                break
        else:
            raise RuntimeError(
                "Please install %s \"csv_reconcile.scorers\" plugin" %
                (plugin,))
    elif len(eps) == 1:
        entrypoint = next(iter(eps))

    if entrypoint is None:
        # print out options
        print(
            "There are several scorers available.  Please choose one of the following with the --scorer option."
        )
        for ep in eps:
            print("  %s" % (ep.name,))
        return None

    entrypoint.load()
    return entrypoint


@click.group()
def cli():
    pass


def doinit(config, scorerOption, csvfile, idcol, namecol):

    app = create_app(config or "", scorerOption=scorerOption or "")
    if app is None:
        return

    with app.app_context():
        initdb.init_db_with_context(csvfile, idcol, namecol)
        click.echo('Initialized the database.')
    return app


@cli.command()
@click.option('--config', help='config file')
@click.option('--scorer', 'scorerOption', help='scoring plugin to use')
@click.argument('csvfile')
@click.argument('idcol')
@click.argument('namecol')
def init(config, scorerOption, csvfile, idcol, namecol):
    return doinit(config, scorerOption, csvfile, idcol, namecol)

@cli.command()
@click.option('--config', help='config file')
@click.option('--scorer', 'scorerOption', help='scoring plugin to use')
@click.option('--init-db', is_flag=True, help='initialize the db')
@click.argument('csvfile')
@click.argument('idcol')
@click.argument('namecol')
def run(config, scorerOption, init_db, csvfile, idcol, namecol):
    print('''
#########################################################
##         WARNING: The interface is deprecated        ##
#########################################################

Please run init once to initialize the database and serve to run the server.
See --help for details.
''')

    app = None
    if init_db:
        app = doinit(config, scorerOption, csvfile, idcol, namecol)

    app = app or create_app(config)
    from werkzeug.serving import WSGIRequestHandler
    WSGIRequestHandler.protocol_version = "HTTP/1.1"
    app.run(debug=False)


@cli.command()
def serve():

    # Config should have been copied during the init phase
    app = create_app()
    from werkzeug.serving import WSGIRequestHandler
    WSGIRequestHandler.protocol_version = "HTTP/1.1"
    app.run(debug=False)


def main():
    nonopts = [a for a in sys.argv if not a.startswith('--')]

    if len(nonopts) > 1 and nonopts[1] not in 'run init serve':
        print('''
#########################################################
##     WARNING: The interface has changed slightly.    ##
#########################################################
Please use one of the subcommands. See --help for details.

''')
    return cli()



from collections import defaultdict
from . import scorer


def reconcileStrings(db, items, **kwargs):
    # Use index as query id
    batch = dict(enumerate({'query': s} for s in items))

    ret = processQueryBatch(db, batch, **kwargs)

    # Return as a list of pairs of query matched to result
    return list(zip(items, (res for _, res in sorted(ret.items()))))


def processQueryBatch(db, batch, limit=None, threshold=0.0, **scoreOptions):
    '''
    Go through db looking for words whose fuzzy match score positively
    '''
    hasFeatures = not getattr(scorer.features, "disabled", False)

    toMatchItems = dict()
    for qid, req in batch.items():
        queryStr = req['query']
        toMatchItems[qid] = scorer.normalizeWord(queryStr, **
                                                 scoreOptions) or queryStr

    # Better to pull these off an sqlite store

    cur = db.cursor()
    normalizedFields = scorer.getNormalizedFields()

    cur.execute('SELECT %s FROM reconcile' %
                (','.join(('word', 'id') + tuple(normalizedFields))))

    picks = defaultdict(list)
    for row in cur:
        compareTo = row[2:] if normalizedFields else row['word']
        if not scorer.valid(compareTo):
            continue

        for qid in batch.keys():
            toMatch = toMatchItems[qid]

            score = scorer.scoreMatch(toMatch, compareTo, **scoreOptions)
            if score is not None and score > threshold:
                picks[qid].append((row, score))

    ret = dict()
    for qid in batch:
        pick = picks[qid]
        lmt = batch[qid].get('limit', limit)
        queryStr = batch[qid]['query']

        res = []
        exacts = []
        cnt = 0
        for row, score in sorted(pick, key=lambda x: -x[1]):
            cnt += 1
            if lmt and cnt > lmt:
                break

            res.append(
                dict(id=row['id'], name=row['word'], score=score, match=False))

            if hasFeatures:
                features = scorer.features(queryStr, row)
                if features is not None:
                    res[-1]['features'] = features

            if res[-1]['name'] == queryStr:
                exacts.append(res[-1])

        # Make match if only one
        if len(res) == 1:
            res[0]['match'] = True
        else:
            if len(exacts) == 1:
                exacts[0]['match'] = True

        # Maybe match if there is a wide gap in score between first match and second?
        ret[qid] = dict(result=res)

    return ret



from flask import current_app
import csv
from chardet.universaldetector import UniversalDetector

from collections import defaultdict
from itertools import count

from .db import get_db, normalizeDBcol

from importlib.resources import read_text
import csv_reconcile
from . import scorer


def initDataTable(db, colnames, idcol):
    cols = []
    cnts = defaultdict(count)
    for col in colnames:
        slug = normalizeDBcol(col)
        slug = f'{slug}{next(cnts[slug])}'
        if col == idcol:
            cols.append('%s TEXT PRIMARY KEY' % (slug,))
        else:
            cols.append('%s TEXT NOT NULL' % (slug,))

        db.execute('INSERT INTO datacols VALUES (?,?,?)',
                   (col, slug, 1 if col == idcol else 0))

    # create data table with the contents of the csv file
    createSQL = 'CREATE TABLE data (\n  %s\n)'
    db.execute(createSQL % (',\n  '.join(cols),))


def initReconcileTable(db, colnames):
    create = [
        'CREATE TABLE reconcile (\n  id TEXT PRIMARY KEY,\n  word TEXT NOT NULL'
    ]
    for col in colnames:
        create.append('%s TEXT NOT NULL' % (col,))

    # create data table with the contents of the csv file
    db.execute(',\n  '.join(create) + '\n)')

def detectEncoding(filenm):
    detector = UniversalDetector()
    for line in open(filenm, 'rb'):
        detector.feed(line)
        if detector.done: break
    detector.close()
    if detector.result['confidence'] > .95:
        return detector.result['encoding']
    return None

def init_db(db,
            csvfilenm,
            idcol,
            searchcol,
            csvencoding=None,
            scoreOptions=None,
            csvkwargs=None):

    enckwarg = dict()
    csvencoding = csvencoding or detectEncoding(csvfilenm)

    if csvencoding:
        enckwarg['encoding'] = csvencoding

    schema = read_text(csv_reconcile, 'schema.sql')
    db.executescript(schema)

    csvkwargs = {} if csvkwargs is None else csvkwargs

    with db:
        # Create a table with ids (as PRIMARY ID), words and bigrams
        with open(csvfilenm, newline='', **enckwarg) as csvfile:
            dialect = None
            try:
                dialect = csv.Sniffer().sniff(csvfile.read(1024))
            except:
                pass

            csvfile.seek(0)
            reader = csv.reader(csvfile, dialect=dialect, **csvkwargs)
            header = next(reader)

            # Throws if col doesn't exist
            searchidx = header.index(searchcol)
            ididx = header.index(idcol)

            normalizedFields = scorer.getNormalizedFields()
            initDataTable(db, header, idcol)
            initReconcileTable(db, normalizedFields)

            datavals = ','.join('?' * len(header))

            for row in reader:
                if len(row) != len(header): continue
                mid = row[ididx]
                word = row[searchidx]
                matchFields = scorer.normalizeRow(word, row, **scoreOptions)
                db.execute(
                    "INSERT INTO reconcile VALUES (%s)" %
                    (','.join('?' * (2 + len(normalizedFields))),),
                    (mid, word) + tuple(matchFields))

                db.execute("INSERT INTO data VALUES (%s)" % (datavals), row)


def init_db_with_context(csvfilenm, idcol, searchcol):
    db = get_db()
    csvkwargs = current_app.config.get('CSVKWARGS', {})
    scoreOptions = current_app.config['SCOREOPTIONS']
    csvencoding = current_app.config.get('CSVENCODING', None)

    return init_db(db,
                   csvfilenm,
                   idcol,
                   searchcol,
                   csvencoding=csvencoding,
                   csvkwargs=csvkwargs,
                   scoreOptions=scoreOptions)



# [[https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient]]
def getDiceCoefficient(bigram1, bigram2):
    '''
    Calculate the Dice coefficient from two normalized sets of bigrams
    '''
    l1 = len(bigram1)
    l2 = len(bigram2)
    i1 = i2 = cnt = 0

    while i1 < l1 and i2 < l2:
        b1 = bigram1[i1:i1 + 2]
        b2 = bigram2[i2:i2 + 2]
        if b1 == b2:
            cnt += 1
            i1 += 2
            i2 += 2
        elif b1 < b2:
            i1 += 2
        else:
            i2 += 2

    # length is twice the number of bigrams
    return 400.0 * cnt / (l1 + l2)



from .db import get_db, getIDCol, getCSVCols

def processDataExtensionBatch(batch):
    ids, props = tuple(batch[x] for x in ('ids', 'properties'))
    names = {p['id'] for p in props}
    cols = {colnm: nm for colnm, nm in getCSVCols() if colnm in names}
    idcol = getIDCol()


    db = get_db()
    cur = db.cursor()
    # Could use some defensiveness in generating this SQL
    sql = "SELECT %s,%s FROM data WHERE %s in (%s)" % (idcol, ','.join(cols.keys()), idcol, ','.join('?' * len(ids)))
    cur.execute(sql, ids)
    rows = dict()
    for row in cur:
        rows[row[idcol]] = {col: [{'str': row[col]}] for col in cols}

    meta = [dict(id=p['id'], name=cols[p['id']]) for p in props]

    return dict(meta=meta, rows=rows)



import os.path
import sqlite3

from flask import current_app, g
from normality import slugify


def normalizeDBcol(col):
    return slugify(col).replace('-', '_')


def getCSVCols():
    cur = get_db().cursor()
    cur.execute("SELECT * FROM datacols")
    return [(row['colname'], row['name']) for row in cur]


def getIDCol():
    cur = get_db().cursor()

    cur.execute("SELECT colname FROM datacols WHERE isid == 1")
    res = cur.fetchall()
    if len(res) != 1:
        raise RuntimeError("database not properly initialized")
    return res[0]['colname']


def get_db():
    if 'db' not in g:
        g.db = sqlite3.connect(os.path.join(current_app.instance_path,
                                            current_app.config['DATABASE']),
                               detect_types=sqlite3.PARSE_DECLTYPES)
        g.db.row_factory = sqlite3.Row

    return g.db


def close_db(e=None):
    db = g.pop('db', None)

    if db is not None:
        db.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['csv_reconcile', 'csv_reconcile_dice']

package_data = \
{'': ['*']}

install_requires = \
['chardet>=4.0.0,<5.0.0',
 'cython>=0.29.21,<0.30.0',
 'flask-cors>=3.0.10,<4.0.0',
 'flask>=2.0,<3.0',
 'normality>=2.1.1,<3.0.0']

extras_require = \
{':python_version < "3.10"': ['importlib_metadata>=4.5.0,<5.0.0']}

entry_points = \
{'console_scripts': ['csv-reconcile = csv_reconcile:main'],
 'csv_reconcile.scorers': ['dice = csv_reconcile_dice']}

setup_kwargs = {
    'name': 'csv-reconcile',
    'version': '0.3.2',
    'description': 'OpenRefine reconciliation service backed by csv resource',
    'long_description': '-   <a href="#csv-reconcile" id="toc-csv-reconcile">CSV Reconcile</a>\n    -   <a href="#quick-start" id="toc-quick-start">Quick start</a>\n    -   <a href="#poetry" id="toc-poetry">Poetry</a>\n        -   <a href="#prerequesites" id="toc-prerequesites">Prerequesites</a>\n        -   <a href="#running" id="toc-running">Running</a>\n        -   <a href="#building" id="toc-building">Building</a>\n    -   <a href="#description" id="toc-description">Description</a>\n    -   <a href="#usage" id="toc-usage">Usage</a>\n        -   <a href="#initialization" id="toc-initialization">Initialization</a>\n        -   <a href="#running-the-service" id="toc-running-the-service">Running the\n            service</a>\n        -   <a href="#deprecated" id="toc-deprecated">Deprecated</a>\n    -   <a href="#common-configuration" id="toc-common-configuration">Common\n        configuration</a>\n    -   <a href="#built-in-preview-service"\n        id="toc-built-in-preview-service">Built-in preview service</a>\n    -   <a href="#scoring-plugins" id="toc-scoring-plugins">Scoring plugins</a>\n        -   <a href="#implementing" id="toc-implementing">Implementing</a>\n        -   <a href="#installing" id="toc-installing">Installing</a>\n        -   <a href="#using" id="toc-using">Using</a>\n        -   <a href="#known-plugins" id="toc-known-plugins">Known plugins</a>\n    -   <a href="#testing" id="toc-testing">Testing</a>\n        -   <a href="#tests-layout" id="toc-tests-layout">Tests layout</a>\n        -   <a href="#running-tests" id="toc-running-tests">Running tests</a>\n    -   <a href="#future-enhancements" id="toc-future-enhancements">Future\n        enhancements</a>\n\n# CSV Reconcile\n\nA [reconciliation service](https://github.com/reconciliation-api/specs)\nfor [OpenRefine](https://openrefine.org/) based on a CSV file similar to\n[reconcile-csv](http://okfnlabs.org/reconcile-csv/). This one is written\nin Python and has some more configurability.\n\n## Quick start\n\n-   Clone this repository\n\n-   Run the service\n\n        $ python -m venv venv                                             # create virtualenv\n        $ venv/bin/pip install csv-reconcile                              # install package\n        $ source venv/bin/activate                                        # activate virtual environment\n        (venv) $ csv-reconcile init sample/reps.tsv item itemLabel        # initialize the service\n        (venv) $ csv-reconcile serve                                      # run the service\n        (venv) $ deactivate                                               # remove virtual environment\n\nThe service is run at <http://127.0.0.1:5000/reconcile>. You can point\nat a different host:port by adding\n[SERVER\\_NAME](https://flask.palletsprojects.com/en/0.12.x/config/) to\nthe sample.cfg. Since this is running from a virtualenv, you can simply\ndelete the whole lot to clean up.\n\nIf you have a C compiler installed you may prefer to install the sdist\n`dist/csv-reconcile-0.1.0.tar.gz` which will build a\n[Cython](https://cython.readthedocs.io/en/latest/) version of the\ncomputationally intensive fuzzy match routine for speed. With `pip` add\nthe option `--no-binary csv-reconcile`.\n\n## Poetry\n\n### Prerequesites\n\nYou\'ll need to have both [poetry](https://python-poetry.org/docs/) and\n[poethepoet](https://pypi.org/project/poethepoet/0.0.3/) installed. For\npublishing to [PyPI](https://pypi.org/) [pandoc](https://pandoc.org/) is\nrequired.\n\n### Running\n\nThis is packaged with [poetry](https://python-poetry.org/docs/), so you\ncan use those commands if you have it installed.\n\n    $ poe install\n    $ poetry run csv-reconcile init sample/reps.tsv item itemLabel\n    $ poetry run csv-reconcile serve\n\n### Building\n\nBecause this package uses a `README.org` file and `pip` requires a\n`README.md`, there are extra build steps beyond what `poetry` supplies.\nThese are managed using\n[poethepoet](https://pypi.org/project/poethepoet/0.0.3/). Thus building\nis done as follows:\n\n    $ poe build\n\nIf you want to build a platform agnostic wheel, you\'ll have to comment\nout the `build =\n    "build.py"` line from `pyproject.toml` until `poetry` supports\n[selecting build\nplatform](https://github.com/python-poetry/poetry/issues/3594).\n\n## Description\n\nThis reconciliation service uses [Dice coefficient\nscoring](https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient)\nto reconcile values against a given column in a\n[CSV](https://en.wikipedia.org/wiki/Comma-separated_values) file. The\nCSV file must contain a column containing distinct values to reconcile\nto. We\'ll call this the *id column*. We\'ll call the column being\nreconciled against the *name column*.\n\nFor performance reasons, the *name column* is preprocessed to normalized\nvalues which are stored in an\n[sqlite](https://www.sqlite.org/index.html) database. This database must\nbe initialized at least once by running the init sub-command. Once\ninitialized this need not be run for subsequent runs.\n\nNote that the service supplies all its data with a dummy *type* so there\nis no reason to reconcile against any particular *type*.\n\nIn addition to reconciling against the *name column*, the service also\nfunctions as a [data extension\nservice](https://reconciliation-api.github.io/specs/latest/#data-extension-service),\nwhich offers any of the other columns of the CSV file.\n\nNote that Dice coefficient scoring is agnostic to word ordering.\n\n## Usage\n\nBasic usage involves two steps:\n\n-   initialization\n-   running the service\n\nInitialization primes the database with the data processed from the CSV\nfile with the `init` subcommand. There are several options for running\nthe service as described below.\n\n### Initialization\n\nBasic usage of the `init` sub-command requires passing the name of the\nCSV file, the *id column* and the *name column*.\n\n    (venv) $ csv-reconcile --help\n    Usage: csv-reconcile [OPTIONS] COMMAND [ARGS]...\n\n    Options:\n      --help  Show this message and exit.\n\n    Commands:\n      init\n      run\n      serve\n    (venv) $ csv-reconcile init --help\n    Usage: csv-reconcile init [OPTIONS] CSVFILE IDCOL NAMECOL\n\n    Options:\n      --config TEXT  config file\n      --scorer TEXT  scoring plugin to use\n      --help         Show this message and exit.\n    (venv) $ poetry run csv-reconcile serve --help\n    Usage: csv-reconcile serve [OPTIONS]\n\n    Options:\n      --help         Show this message and exit.\n    (venv) $\n\nThe `--config` option is used to point to a configuration file. The file\nis a [Flask\nconfiguration](https://flask.palletsprojects.com/en/1.1.x/config/) and\nhence is Python code though most configuration is simply setting\nvariables to constant values.\n\n### Running the service\n\nThe simplest way to run the service is to use Flask\'s built-in web\nserver with the `serve` subcommand which takes no arguments. However, as\nmentioned in the [Flask\ndocumentation](https://flask.palletsprojects.com/en/2.0.x/deploying/),\nthis server is not suitable for production purposes.\n\nFor a more hardened service, you can use one of the other deployment\noptions mentioned in that documentation. For example, gunicorn can be\nrun as follows:\n\n    (venv) $ gunicorn -w 4 \'csv_reconcile:create_app()\'\n    1-11-16 17:40:20 +0900] [84625] [INFO] Starting gunicorn 20.1.0\n    1-11-16 17:40:20 +0900] [84625] [INFO] Listening at: http://127.0.0.1:8000 (84625)\n    1-11-16 17:40:20 +0900] [84625] [INFO] Using worker: sync\n    1-11-16 17:40:20 +0900] [84626] [INFO] Booting worker with pid: 84626\n    1-11-16 17:40:20 +0900] [84627] [INFO] Booting worker with pid: 84627\n    1-11-16 17:40:20 +0900] [84628] [INFO] Booting worker with pid: 84628\n    1-11-16 17:40:20 +0900] [84629] [INFO] Booting worker with pid: 84629\n    ...\n\nOne thing to watch out for is that the default manifest points the\nextension service to port 5000, the default port for the Flask built-in\nweb server. If you want to use the extension service when deploying to a\ndifferent port, you\'ll want to be sure to override that part of the\nmanifest in your config file. You\'ll need something like the following:\n\n    MANIFEST = {\n        "extend": {\n            "propose_properties": {\n                "service_url": "http://localhost:8000",\n                "service_path": "/properties"\n            }\n        }\n    }\n\nNote also that the configuration is saved during the `init` step. If you\nchange the config, you\'ll need to re-run that step. You may also need to\ndelete and re-add the service in OpenRefine.\n\n### Deprecated\n\nThe `run` subcommand mimics the old behavior which combined the\ninitialization step with the running of the service. This may be removed\nin a future release.\n\n## Common configuration\n\n-   `SERVER_NAME` - The host and port the service is bound to. e.g.\n    `SERVER_NAME=localhost:5555`. ( Default localhost:5000 )\n-   `CSVKWARGS` - Arguments to pass to\n    [csv.reader](https://docs.python.org/3/library/csv.html). e.g.\n    `CSVKWARGS={\'delimiter\': \',\', \'quotechar\': \'"\'}` for comma delimited\n    files using `"` as quote character.\n-   `CSVENCODING` - Encoding of the CSV file. e.g.\n    `CSVENCODING="utf-8-sig"` is the encoding used for data downloaded\n    from\n    [GNIS](https://www.usgs.gov/core-science-systems/ngp/board-on-geographic-names/download-gnis-data).\n-   `SCOREOPTIONS` - Options passed to scoring plugin during\n    normalization. e.g.\n    `SCOREOPTIONS={\'stopwords\':[\'lake\',\'reservoir\']}`\n-   `LIMIT` - The maximum number of reonciliation candidates returned\n    per entry. ( Default 10 ) e.g. `LIMIT=10`\n-   `THRESHOLD` - The minimum score for returned reconciliation\n    candidates. ( Default 30.0 ) e.g. `THRESHOLD=80.5`\n-   `DATABASE` - The name of the generated sqlite database containing\n    pre-processed values. (Default `csvreconcile.db`) e.g.\n    `DATABASE=\'lakes.db\'` You may want to change the name of the\n    database if you regularly switch between databases being used.\n-   `MANIFEST` - Overrides for the service manifest. e.g.\n    `MANIFEST={"name": "My service"}` sets the name of the service to\n    "My service".\n\nThis last is most interesting. If your data is coming from\n[Wikidata](https://www.wikidata.org) and your *id column* contains [Q\nvalues](https://www.wikidata.org/wiki/Help:Items), then a manifest like\nthe following will allow your links to be clickable inside OpenRefine.\n\n    MANIFEST = {\n      "identifierSpace": "http://www.wikidata.org/entity/",\n      "schemaSpace": "http://www.wikidata.org/prop/direct/",\n      "view": {"url":"https://www.wikidata.org/wiki/{{id}}"},\n      "name": "My reconciliation service"\n    }\n\nIf your CSV is made up of data taken from another [reconciliation\nservice](https://reconciliation-api.github.io/testbench/), you may\nsimiliarly copy parts of their manifest to make use of their features,\nsuch as the [preview\nservice](https://reconciliation-api.github.io/specs/latest/#preview-service).\nSee the reconciliation spec for details.\n\n## Built-in preview service\n\nThere is a preview service built into the tool. (Thanks\n[b2m](https://github.com/b2m)!) You can turn it on by adding the\nfollowing to your manifest:\n\n    "preview": {\n       "url": "http://localhost:5000/preview/{{id}}",\n       "width": 400,\n       "height": 300\n    }\n\nNote that if you reconcile against a service with a preview service\nenabled, a link to the service becomes part of the project. Thus if you\nbring the service down, your project will have hover over pop-ups to an\nunavailable service. One way around this is to copy the `recon.match.id`\nto a new column which can be re-reconciled to the column by id if you\nbring the service back up again whether or not you have preview service\nenabled. (Perhaps OpenRefine could be smarter about enabling this\npop-ups only when the service is active.)\n\n## Scoring plugins\n\nAs mentioned above the default scoring method is to use [Dice\ncoefficient\nscoring](https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient),\nbut this method can be overridden by implementing a\n`cvs_reconcile.scorers` plugin.\n\n### Implementing\n\nA plugin module may override any of the methods in the\n`csv_reconcile.scorers` module by simply implementing a method of the\nsame name with the decorator `@cvs_reconcile.scorer.register`.\n\nSee `csv_reconcile_dice` for how Dice coefficient scoring is\nimplemented.\n\nThe basic hooks are as follows:\n\n-   `normalizedWord(word, **scoreOptions)` preprocesses values to be\n    reconciled to produce a tuple used in fuzzy match scoring. The value\n    of `SCOREOPTIONS` in the configuration will be passed in to allow\n    configuration of this preprocessing. This hook is required.\n-   `normalizedRow(word, row, **scoreOptions)` preprocesses values to be\n    reconciled against to produce a tuple used in fuzzy match scoring.\n    Note that both the reconciled column and the entire row is available\n    for calculating the normalized value and that the column reconciled\n    against is required even when not used. The value of `SCOREOPTIONS`\n    in the configuration will be passed in to allow configuration of\n    this preprocessing. This defaults to calling\n    normalizeWord(word,\\*\\*scoreOptions).\n-   `getNormalizedFields()` returns a tuple of names for the columns\n    produced by `normalizeWord()`. The length of the return value from\n    both functions must match. This defaults to calling\n    normalizeWord(word,\\*\\*scoreOptions). This hook is required.\n-   `processScoreOptions(options)` is passed the value of `SCOREOPTIONS`\n    to allow it to be adjusted prior to being used. This can be used for\n    adding defaults and/or validating the configuration. This hook is\n    optional\n-   `scoreMatch(left, right, **scoreOptions)` gets passed two tuples as\n    returned by `normalizedWord()`. The `left` value is the value being\n    reconciled and the `right` value is the value being reconciled\n    against. The value of `SCOREOPTIONS` in the configuration will be\n    passed in to allow configuration of this preprocessing. Returning a\n    score of `None` will not add tested value as a candidate. This hook\n    is required.\n-   `valid(normalizedFields)` is passed the normalized tuple prior to\n    being scored to make sure it\'s appropriate for the calculation. This\n    hook is optional.\n-   `features(word, row, **scoreOptions)` calculates\n    [features](https://reconciliation-api.github.io/specs/latest/#reconciliation-query-responses)\n    using the query string and the normalized row. By default\n    calculating features is disabled. Implementions of this hook are\n    automatically enabled. This hook is optional.\n\n### Installing\n\nHooks are automatically discovered as long as they provide a\n`csv_reconcile.scorers` [setuptools entry\npoint](https://setuptools.readthedocs.io/en/latest/userguide/entry_point.html).\nPoetry supplies a\n[plugins](https://python-poetry.org/docs/pyproject/#plugins)\nconfiguration which wraps the setuptools funtionality.\n\nThe default Dice coefficent scoring is supplied via the following\nsnippet from `pyproject.toml` file.\n\n    [tool.poetry.plugins."csv_reconcile.scorers"]\n    "dice" = "csv_reconcile_dice"\n\nHere `dice` becomes the name of the scoring option and\n`csv_reconcile_dice` is the package implementing the plugin.\n\n### Using\n\nIf there is only one scoring plugin available, that plugin is used. If\nthere are more than one available, you will be prompted to pass the\n`--scorer` option to select among the scoring options.\n\n### Known plugins\n\nSee [wiki](https://github.com/gitonthescene/csv-reconcile/wiki) for list\nof known plugins.\n\n## Testing\n\nThough I long for the old days when a unit test was a unit test, these\ndays things are a bit more complicated with various versions of `Python`\nand installation of plugins to manage. Now we have to wrestle with\n[virtual environments](https://docs.python.org/3/tutorial/venv.html).\n`poetry` handles the virtual environment for developing, but testing\ninvolves covering more options.\n\n### Tests layout\n\nThe tests directory structure is the following:\n\n    tests\n        main\n        plugins\n            geo\n\nTests for the main package are found under `main` and don\'t require\ninstalling any other packages whereas tests under `plugins` require the\ninstallation of the given plugin.\n\n### Running tests\n\n1.  Basic tests\n\n    These tests are written with\n    [pytest](https://docs.pytest.org/en/6.2.x/contents.html) and can be\n    running through `poetry` as follows:\n\n        $ poetry run pytest\n\n    To avoid the complications that come from installing plugins, there\n    is a `poe` script for running only the tests under main which can be\n    invoked as follows:\n\n        $ poe test\n\n    For steady state developing this is probably the command you\'ll use\n    most often.\n\n2.  Build matrices\n\n    The GitHub Actions for this project currently use a [build\n    matrix](https://docs.github.com/en/actions/learn-github-actions/managing-complex-workflows#using-a-build-matrix)\n    across a couple of architectures and several versions of `Python`,\n    but a similar effect can be achieved using\n    [nox](https://nox.thea.codes/en/stable/tutorial.html).\n\n    `nox` manages the creation of various virtual environments in what\n    they call "sessions", from which various commands can be run. This\n    project\'s `noxfile.py` manages the installation of the\n    `csv-reconcile-geo` plugin for the plugin tests as well as running\n    across several versions of `Python`. See the `nox` documentation for\n    detail.\n\n    Some versions of this command you\'re likely to run are as follows:\n\n        $ nox      # Run all the tests building virtual environemnts from scratch\n        $ nox -r   # Reuse previously built virtual environments for speed\n        $ nox -s test_geo  # Run only the tests for the csv-reconcile-geo plugin\n        $ nox -s test_main -p 3.8   # Run only the main tests with Python3.8\n\n    Eventually, the GitHub Actions may be changed to use\n    [setup-nox](https://github.com/marketplace/actions/setup-nox).\n\n## Future enhancements\n\nIt would be nice to add support for using\n[properties](https://reconciliation-api.github.io/specs/latest/#structure-of-a-reconciliation-query)\nas part of the scoring, so that more than one column of the csv could be\ntaken into consideration.\n',
    'author': 'Douglas Mennella',
    'author_email': 'trx2358-pypi@yahoo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gitonthescene/csv-reconcile',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from .db import get_db, getCSVCols, getIDCol


def getEntity(entity_id):
    id_col = getIDCol()
    cols = dict(getCSVCols())

    cur = get_db().cursor()
    cur.execute(f"SELECT * FROM data WHERE {id_col}=? LIMIT 1", (entity_id,))
    row = cur.fetchone()
    if not row:
        return None
    return {cols[col]: value for col, value in zip(row.keys(), row)}
